from typing import Any, Dict, Optional, Union

from grai_schemas.generics import MalformedMetadata, Metadata
from grai_schemas.utilities import merge
from grai_schemas.v1.metadata.edges import BaseEdgeMetadataV1
from grai_schemas.v1.metadata.nodes import BaseNodeMetadataV1

MetadataType = Union[BaseNodeMetadataV1, BaseEdgeMetadataV1]


@merge.register
def merge_malformed_left(metadata: MalformedMetadata, other_metadata: Any) -> Metadata:
    """ """
    return other_metadata


@merge.register
def merge_malformed_right(metadata: Any, other_metadata: MalformedMetadata):
    """ """
    raise ValueError(
        "Cannot merge malformed metadata into valid metadata. If you received this error check if any of the metadata"
        "in your objects is of type `MalformedMetadata`. This could indicate a bad value has been stored in your"
        "instance of Grai."
    )


def merge_tags(a: Optional[list], b: Optional[list]) -> list:
    a_tag = set(a) if a is not None else set()
    b_tag = set(b) if b is not None else set()
    return list(a_tag.union(b_tag))


@merge.register
def merge_grai_node_v1_metadata(metadata: BaseNodeMetadataV1, other_metadata: BaseNodeMetadataV1) -> BaseNodeMetadataV1:
    """ """
    new_metadata = merge(dict(metadata), dict(other_metadata))
    new_metadata["tags"] = merge_tags(metadata.tags, other_metadata.tags)
    return BaseNodeMetadataV1(**new_metadata)


@merge.register
def merge_grai_edge_v1_metadata(metadata: BaseEdgeMetadataV1, other_metadata: BaseEdgeMetadataV1) -> BaseEdgeMetadataV1:
    """ """
    new_metadata = merge(dict(metadata), dict(other_metadata))
    new_metadata["tags"] = merge_tags(metadata.tags, other_metadata.tags)
    return BaseEdgeMetadataV1(**new_metadata)



from typing import Dict, Literal, Union

from grai_schemas.base import GraiType
from grai_schemas.generics import GraiBaseModel


class Schema(GraiBaseModel):
    """ """

    entity: GraiType

    @classmethod
    def to_model(cls, item: Dict, version: Literal["v1"], typing_type: Literal["Node", "Edge"]) -> GraiType:
        """

        Args:
            item (Dict):
            version (Literal["v1"]):
            typing_type (Literal["Node", "Edge"]):

        Returns:

        Raises:

        """
        result = {
            "type": typing_type,
            "version": version,
            "spec": item,
        }
        return cls(entity=result).entity



from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from grai_schemas.generics import GraiBaseModel, MalformedMetadata
from grai_schemas.v1.metadata.generics import GenericAttributes


class EdgeMetadataTypeLabels(Enum):
    """ """

    generic = "Generic"
    table_to_column = "TableToColumn"
    column_to_column = "ColumnToColumn"
    table_to_table = "TableToTable"


EdgeTypeLabelLiterals = Literal["Generic", "TableToColumn", "ColumnToColumn", "TableToTable"]


class BaseEdgeMetadataV1(GraiBaseModel):
    """ """

    type: Literal["EdgeV1"] = "EdgeV1"
    version: Literal["v1"] = "v1"
    edge_type: EdgeTypeLabelLiterals
    edge_attributes: GenericAttributes
    tags: Optional[List[str]]


class MalformedEdgeMetadataV1(MalformedMetadata, BaseEdgeMetadataV1):
    edge_type: Optional[str] = "Malformed"  # type: ignore
    edge_attributes: Optional[Any] = GenericAttributes()  # type: ignore


class GenericEdgeMetadataV1(BaseEdgeMetadataV1):
    edge_type: Literal["Generic"]
    edge_attributes: GenericAttributes = GenericAttributes()


class TableToColumnAttributes(GenericAttributes):
    """ """

    version: Literal["v1"] = "v1"
    pass


class TableToColumnMetadata(BaseEdgeMetadataV1):
    """ """

    edge_type: Literal["TableToColumn"]
    edge_attributes: TableToColumnAttributes = TableToColumnAttributes()


class TableToTableAttributes(GenericAttributes):
    """ """

    version: Literal["v1"] = "v1"
    pass


class TableToTableMetadata(BaseEdgeMetadataV1):
    """ """

    edge_type: Literal["TableToTable"]
    edge_attributes: TableToTableAttributes = TableToTableAttributes()


class ColumnToColumnAttributes(GenericAttributes):
    """ """

    version: Literal["v1"] = "v1"
    preserves_data_type: Optional[bool] = None
    preserves_nullable: Optional[bool] = None
    preserves_unique: Optional[bool] = None


class ColumnToColumnMetadata(BaseEdgeMetadataV1):
    """ """

    edge_type: Literal["ColumnToColumn"]
    edge_attributes: ColumnToColumnAttributes = ColumnToColumnAttributes()


Metadata = Union[TableToColumnMetadata, TableToTableMetadata, ColumnToColumnMetadata, GenericEdgeMetadataV1]



import pathlib
from typing import Any

import orjson
from pydantic.json import pydantic_encoder


def orjson_defaults(obj: Any) -> Any:
    """

    Args:
        obj (Any):

    Returns:

    Raises:

    """
    if isinstance(obj, set):
        return list(obj)
    elif isinstance(obj, (pathlib.PosixPath, pathlib.WindowsPath)):
        return str(obj)
    else:
        raise Exception(f"No supported JSON serialization format for objects of type {type(obj)}")


def dump_json(v, *, default):
    return orjson.dumps(v, default=default, option=orjson.OPT_NON_STR_KEYS).decode()


def load_json(*args, **kwargs):
    return orjson.loads(*args, **kwargs)



from typing import Callable, Dict, List, Literal, Optional, Type, Union
from uuid import UUID

from grai_schemas.v1.generics import GraiBaseModel, NamedID, UuidID
from grai_schemas.v1.metadata.edges import GenericEdgeMetadataV1, Metadata
from grai_schemas.v1.metadata.metadata import EdgeMetadataV1, GraiEdgeMetadataV1
from grai_schemas.v1.node import NodeIdTypes
from grai_schemas.v1.source import DataSourceMixin, DataSourcesMixin
from pydantic import validator


class EdgeNamedID(NamedID):
    """ """

    pass


class EdgeUuidID(UuidID):
    """ """

    pass


EdgeIdTypes = Union[EdgeUuidID, EdgeNamedID]


class BaseSpec(GraiBaseModel):
    """ """

    display_name: Optional[str]
    source: NodeIdTypes
    destination: NodeIdTypes
    is_active: Optional[bool] = True
    workspace: Optional[UUID]

    def __str__(self):
        return f"Edge[Node({self.source}) -> Node({self.destination})]"


class SourcedEdgeSpecMetadataMixin(GraiBaseModel):
    metadata: GraiEdgeMetadataV1 = GraiEdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"))

    @validator("metadata", always=True, pre=True)
    def validate_metadata(cls, v: Optional[Union[Dict, GraiEdgeMetadataV1]]) -> GraiEdgeMetadataV1:
        if isinstance(v, GraiEdgeMetadataV1):
            return v
        elif isinstance(v, dict):
            v.setdefault("grai", GenericEdgeMetadataV1(edge_type="Generic"))
            return GraiEdgeMetadataV1(**v)
        elif v is None:
            return GraiEdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"))
        raise ValueError(f"Invalid metadata: {v}. Expected either None, a dict, or a MetadataV1 instance.")


class NamedSourceSpec(EdgeNamedID, BaseSpec, SourcedEdgeSpecMetadataMixin, DataSourceMixin):
    """ """

    pass


class IDSourceSpec(EdgeUuidID, BaseSpec, SourcedEdgeSpecMetadataMixin, DataSourceMixin):
    """ """

    pass


SourcedEdgeSpec = Union[IDSourceSpec, NamedSourceSpec]


class SourcedEdgeV1(GraiBaseModel):
    type: Literal["SourceEdge"]
    version: Literal["v1"]
    spec: SourcedEdgeSpec

    @classmethod
    def from_spec(cls, spec_dict: Dict) -> "SourcedEdgeV1":
        """

        Args:
            spec_dict (Dict):

        Returns:

        Raises:

        """
        return cls(version="v1", type="SourceEdge", spec=spec_dict)

    def __hash__(self):
        return hash(self.spec)


class EdgeSpecMetadataMixin(GraiBaseModel):
    metadata: EdgeMetadataV1 = EdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"), sources={})

    @validator("metadata", always=True, pre=True)
    def validate_metadata(cls, v: Optional[Union[Dict, EdgeMetadataV1]]) -> EdgeMetadataV1:
        if isinstance(v, EdgeMetadataV1):
            return v
        elif isinstance(v, dict):
            v.setdefault("grai", GenericEdgeMetadataV1(edge_type="Generic"))
            v.setdefault("sources", {})
            return EdgeMetadataV1(**v)
        elif v is None:
            return EdgeMetadataV1(grai=GenericEdgeMetadataV1(edge_type="Generic"), sources={})
        raise ValueError(f"Invalid metadata: {v}. Expected either None, a dict, or a MetadataV1 instance.")


class NamedSpec(EdgeNamedID, BaseSpec, EdgeSpecMetadataMixin, DataSourcesMixin):
    """ """

    pass


class IDSpec(EdgeUuidID, BaseSpec, EdgeSpecMetadataMixin, DataSourcesMixin):
    """ """

    pass


EdgeSpec = Union[IDSpec, NamedSpec]


class EdgeV1(GraiBaseModel):
    """ """

    type: Literal["Edge"]
    version: Literal["v1"]
    spec: EdgeSpec

    @classmethod
    def from_spec(cls, spec_dict: Dict) -> "EdgeV1":
        """

        Args:
            spec_dict (Dict):

        Returns:

        Raises:

        """
        return cls(version="v1", type="Edge", spec=spec_dict)

    def __hash__(self):
        return hash(self.spec)
